#Author: Zeynep G. Saribatur
#
#
#Given as inputs: a program + a sequence of atoms to be omitted (or a file containing the sequence of atoms), 
# step 1: constructs the abstract program, 
# step 2: gets an abstract answer set, 
# step 3: checks the correctness of the abstract answer set, 
# step 4: in case of spuriousness, it uses the answer set to debug the abstraction and decides on a refinement (i.e., the atoms to be added back)
# step 5: refines the abstraction, and moves to step 1.
# this loop continues until the abstraction is refined all the way down (i.e., all atoms are added back) or a concrete answer set is found (or unsatisfiability is achieved)
#
#Uses the files in auxiliary folder
#Creates in outputs folder: *_finalomit_info, *_finalomit_absprogram, *_finalomittedatoms, *_progatoms
#
#IMPORTANT: the input program should be grounded (with no parentheses or commas) and each rule should contain rulenumbers 
#- please use the script grounding_to_groundatoms.sh to convert your ground program into the desired input format
#this is required for forwarding the input to auxiliary/spock.jar


import re
import pexpect
import time
import sys
import csv
import string
import os



def createOmittedProgram(fileorg,omitAtom):
	
	change=False
	omit=False
	
	changedRules=[]

	fin=open(fileorg,"r")
	textline = fin.readline()

	global atoms

	fout = open("outputs/org_program_dg.lp","w")

	fout2 = open("outputs/omitted_program_info.lp","w")
	for i in range(len(omitAtom)):
		fout2.write("omittedAtom("+omitAtom[i]+").")
	fout2.write("\n")

	newText=""
	while textline:
		ruleName=""
		ruleHead=""
		ruleBody=""
		if textline[0]=="%":
			textline = fin.readline()
			continue
		line=string.replace(textline,"[^'not'] ","")
		if line=="\n":
			textline = fin.readline()
			continue
		#r1: c :- not cc.
		ruleName=re.findall(r"(.*):.*:-.*.",line)
		if ruleName!=[]:
			ruleName=ruleName[0]
			ruleHead=re.findall(r".*:(.*):-.*.",line)[0]
			ruleBody=re.findall(r".*:.*:-(.*).",line)[0]
		else:
			ruleName=re.findall(r"(.*):.*.",line)[0]
			ruleHead=re.findall(r".*:(.*).",line)[0]
			ruleBody=[]
			

		if "i" in ruleName:
			textline = fin.readline()
			continue
		
		constraint=False
		fout.write("rule("+ruleName+").\n")
		if re.search('[a-zA-Z]', ruleHead)!=None:
			#TODO need to consider the possibility of choice rules	
			ruleHeadAtom=ruleHead.replace("}","")
			ruleHeadAtom=ruleHeadAtom.replace("{","")
			fout.write("head("+ruleName+","+ruleHeadAtom+").\n")
			if ruleHeadAtom not in atoms:
				atoms.append(ruleHeadAtom)
		else:
			constraint=True

		
				
		if ruleBody!=[]:
			ruleBody=ruleBody.split(",")
			ruleNegBody=[string.replace(s,"not ","") for s in ruleBody if "not " in s]
			rulePosBody=[s for s in ruleBody if "not " not in s]
			for i in range(len(ruleNegBody)):
				fout.write("negBody("+ruleName+","+ruleNegBody[i]+").\n")
			for i in range(len(rulePosBody)):
				fout.write("posBody("+ruleName+","+rulePosBody[i]+").\n")
		else:
			ruleNegBody=[]
			rulePosBody=[]
		ruleHead=ruleHead.replace(" ","")
		for a in omitAtom:
			if a == ruleHead or "{"+a+"}" == ruleHead:
				omit=True
				fout2.write("omitted("+ruleName+").\n")
			else:
				if omit==True:
					continue
				for i in range(len(ruleNegBody)):
					ruleNegBody[i]=ruleNegBody[i].replace(" ","")
					#TODO this "=" check wouldnt work for non-ground atoms since "p != p(X)"
					if a==ruleNegBody[i]:
						if constraint==True:
							fout2.write("omitted("+ruleName+").\n")
							fout2.write("ko("+ruleName+").\nab_p("+ruleName+") :- ap("+ruleName+").\n")
							omit=True
							break	
						change=True
						ruleNegBody[i]=""
				for i in range(len(rulePosBody)):
					rulePosBody[i]=rulePosBody[i].replace(" ","")
					if a==rulePosBody[i]:
						if constraint==True:
							fout2.write("omitted("+ruleName+").\n")
							fout2.write("ko("+ruleName+").\nab_p("+ruleName+") :- ap("+ruleName+").\n")
							omit=True
							break	
						change=True
						rulePosBody[i]=""
		if change==True and omit!=True:
			fout2.write("changed("+ruleName+").\n")
			changedRules.append(ruleName)

		ruleNegBody=["not "+s for s in ruleNegBody if s!=""]
		rulePosBody=[s for s in rulePosBody if s!=""]
		newBody=""
		newHead=""
		newRule=""
		if change==True and omit!=True and ruleHead!="":
			newBody= ','.join(rulePosBody) 
			if rulePosBody!=[] and ruleNegBody!=[]:
				newBody=newBody+","+ ','.join(ruleNegBody)
			if rulePosBody==[] and ruleNegBody!=[]:
				newBody=','.join(ruleNegBody)
			if "{" not in ruleHead and ruleHead!="":
				newHead="{"+ruleHead+"}"
			if newBody!="":
				newRule=newHead+":-"+newBody+"."
			else:
				newRule=newHead+"."
			newText=newText+newRule+"\n"
		if change==False and omit==False:
			newRule=textline[textline.find(":")+2:len(textline)-1]
			newText=newText+newRule+"\n"
		change=False
		omit=False
		fout.write("\n")
		textline = fin.readline()
	
	fin.close()

	fout.close()
	fout2.close()

	fout = open("outputs/omitted_program.lp","w")
	fout.write(newText)
	fout.close()

	fout = open("outputs/aux.lp","w")
	text=""
	for a in atoms:
		text=text+"holds("+a+") :- "+a+".\n"
	text=text+"atomnum(1.."+str(len(atoms))+").\n"
	fout.write(text)
	fout.close()


	return changedRules


def getSpockOutput(fileorg,omitAtom,changedRules):
	newfilename=fileorg+"_expoexco"
	rules=""	
	if changedRules!=[]:
		rules=','.join(changedRules)
	print "calling: java -jar auxiliary/spock.jar -k -expo -exco -exlo -exrules:"+rules+" -sm "+fileorg+" > "+newfilename
	os.system("java -jar auxiliary/spock.jar -k -expo -exco -exlo -exrules:"+rules+" -sm "+fileorg+" > "+newfilename)
	#TODO spock can not deal with choice rules in the head correctly:
	#	{a}. becomes {a}. ab_c(a) :- a.
	fin=open(newfilename,"r")
	data=fin.readlines()
	fin.close()
	modifyAbCForChoice=[]
	addBlForRule=[]
	for i in range(len(data)):
		for a in omitAtom:
			if "ab_c("+a+") :-" in data[i] :
				data[i]=''
				data[i-1]=''
		if "not not " in data[i]:
			#bl(r1) :- ok(r1), not not d.
			notnotLine = re.findall(r"(.*, not )(not (.*)).",data[i])
			data[i]=notnotLine[0][0]+"aux_"+notnotLine[0][2]+".\n"
			data.append("aux_"+notnotLine[0][2]+":-"+notnotLine[0][1]+".\n")
		catchChoiceFact=re.findall(r"\{(\S+)\}\ *:-\ *ap\((\S+)\),\ *not ko",data[i])
		if catchChoiceFact!=[] and "bl" not in data[i+2]:
			if catchChoiceFact[0][0] not in modifyAbCForChoice: #being aware of cases where same atom can occur more than once
				modifyAbCForChoice.append(catchChoiceFact[0][0])
				addBlForRule.append(catchChoiceFact[0][1])
		for a in modifyAbCForChoice:
			if "ab_c("+a+") :-" in data[i]:
				data[i]=data[i][0:len(data[i])-2]+", bl("+addBlForRule[modifyAbCForChoice.index(a)]+").\n"
				if ":-" in data[i-1]:
					data[i-1]=data[i-1][0:len(data[i-1])-2]+", bl("+addBlForRule[modifyAbCForChoice.index(a)]+").\n"
				else:
					data[i-1]=data[i-1][0:len(data[i-1])-2]+":- bl("+addBlForRule[modifyAbCForChoice.index(a)]+").\n"
		
	fout=open(newfilename,"w")
	fout.writelines(data)
	fout.close()

def createConstraints(ans,omitAtom):
	global atoms

	fout=open("outputs/interpretation.lp","w")
	text=""
	for i in ans:
		i=i.replace(" ","")
		if i=="":
			print "moving on"
		else:
			text=text+":- not "+i+".\n"
	for i in atoms:		
		i=i.replace(" ","")
		if i not in ans and i not in omitAtom:
			text=text+":- "+i+".\n"
	#print text
	fout.write(text)
	fout.close()

	return text

def callDebugger(fileorg):
	fout=open("output_debug.txt","w")
	print "clingo "+fileorg+"_expoexco outputs/omitted_program_info.lp outputs/org_program_dg.lp auxiliary/debug_limitbadomit.lp outputs/aux.lp outputs/interpretation.lp auxiliary/loop_detection.lp"
	p = pexpect.spawn("clingo "+fileorg+"_expoexco outputs/omitted_program_info.lp outputs/org_program_dg.lp auxiliary/debug_limitbadomit.lp outputs/aux.lp outputs/interpretation.lp auxiliary/loop_detection.lp",timeout=600)
#	print "clingo "+fileorg+"_expoexco omitted_program_info.lp org_program_dg.lp debug.lp aux.lp interpretation.lp spock/loop_detection.lp"
#	p = pexpect.spawn("clingo "+fileorg+"_expoexco omitted_program_info.lp org_program_dg.lp debug.lp aux.lp interpretation.lp spock/loop_detection.lp",timeout=600)
	p.logfile=fout
	a=p.expect([pexpect.EOF,'ERROR',pexpect.TIMEOUT])
	fout.close()
	if a == 1 or a==2: # Timeout
		print "Timeout or Error"
		sys.exit()
	fin = open("output_debug.txt","r")
	text = fin.read()
	fin.close()	
	totalmodel=re.findall(r"Models\ +: (\d+)",text)[0]
	print "totalmodel: "+str(totalmodel)
	times=re.findall(r"Time\ +: (\S+)s \(Solving: (\S+)s 1st Model",text)
	print times
	if "UNSAT" in text:
		print "unsat: something's wrong"
		return -1
	else:
		#check if all answer sets have some "badomit"
		#     if not, the interpretation has a concrete answer set!!!!
		for i in range(int(totalmodel)):
			if i<int(totalmodel)-1:
				if "OPTIMUM FOUND" in text:
					test=text[text.find("Answer: "+totalmodel):text.find("OPTIMUM FOUND")]
				else:
					test=text[text.find("Answer: "+str(i+1)):text.find("Answer: "+str(i+2))]
			else:
				test=text[text.find("Answer: "+str(i+1)):text.find("SAT")]
				if test==[]:
					test=text[text.find("Answer: "+str(i+1)):text.find("OPTIMUM FOUND")]
			if "badomit" in test:
				badomits=re.findall(r"badomit\((\S+)\)",test)
				badomitAtoms=[s for s in badomits if "type" not in s]
				return badomitAtoms
			else:
				print "---no bad omit - a concrete answer set??----"
				if "ab_p" in test or "ab_c" in test:
					return -1
				else:
					return 1

def checkAbsProg():
	global unsat
	fout=open("outputs/output_check.txt","w")
	print "clingo outputs/omitted_program.lp"
	p = pexpect.spawn("clingo outputs/omitted_program.lp",timeout=600)
	p.logfile=fout
	a=p.expect([pexpect.EOF,'ERROR',pexpect.TIMEOUT])
	fout.close()
	if a == 1 or a==2: # Timeout
		print "Timeout or Error"
		sys.exit()
	fin = open("outputs/output_check.txt","r")
	text = fin.read()
	fin.close()		
	totalmodel=re.findall(r"Models\ +: (\d+)",text)[0]
	times=re.findall(r"Time\ +: (\S+)s \(Solving: (\S+)s 1st Model",text)
	print times
	ansAll=[]
	if "UNSAT" in text:
		print "unsat: the desired property does not exist in the original program!!"
		unsat=True
	else:
		print "total model: "+totalmodel
		print "sat: need to check the answer sets for concreteness"
		for i in range(int(totalmodel)):
			if i<int(totalmodel)-1:
				test=text[text.find("Answer: "+str(i+1)):text.find("Answer: "+str(i+2))]
			else:
				test=text[text.find("Answer: "+str(i+1)):text.find("SAT")]
			ans=re.findall(r"Answer.*\n(.*)\s*\r",test)
			#check in case there is empty answerset
			if ans!=[]:
				ans=ans[0].split(" ")
			ansAll.append(ans)
	#print ansAll
	return ansAll

def checkSpuriousExists(ansAll,spuriousAns):
	for ans in ansAll:
		count=0
		for a in spuriousAns:
			if a in ans:
				count=count+1
		if count==len(spuriousAns):
			return ans
	return 0

def checkIndeedConcrete(fileorg):
	filey=fileorg.replace("_rulename","")	
	fout = open("outputs/output_concretecheck.txt","w")
	print "clingo "+filey+" outputs/interpretation.lp"
	p = pexpect.spawn("clingo "+filey+" outputs/interpretation.lp",timeout=600)
	p.logfile=fout
	a=p.expect([pexpect.EOF,'ERROR',pexpect.TIMEOUT])
	fout.close()
	if a == 1 or a==2: # Timeout
		print "Timeout or Error"
		sys.exit()
	fin = open("outputs/output_concretecheck.txt","r")
	text = fin.read()
	fin.close()	
	if "UNSAT" in text:
		print "UNSAT! something's wrong!"


fileorg=sys.argv[1]
omitAtom=sys.argv[2]
if "." in omitAtom:
	omitAtomFile=omitAtom
	fin=open(omitAtom,"r")
	omitAtom=fin.read()
	fin.close()
	omitAtom=omitAtom.replace("\n","")
else:
	omitAtomFile=fileorg
omitAtom=omitAtom.split(",")
omitAtomInit=omitAtom
print omitAtom

if os.path.isdir("outputs")==False:
	os.makedirs("outputs")



printatoms=True


atoms=[]
addedBack=[]
refineStep=0
unsat=False

while True:
	changedRules = createOmittedProgram(fileorg,omitAtom)

	getSpockOutput(fileorg,omitAtom,changedRules)
	spockfilename=fileorg+"_expoexco"

	ansAll=checkAbsProg()
	
	spuriousAns=[]
	badomitAtoms=[]
	for ans in ansAll:
		#interpretation=["not c","a","d"]
		print "Create constraints for abs answer " + ','.join(ans)
		interpretation=createConstraints(ans,omitAtom)
		print "Call Debugger"
		res=callDebugger(fileorg)
			
		if res==-1:
			print "--SOMETHING IS WRONG--"
			sys.exit()
		else:
			if res==1:
				print "--CONCRETE ANSWER SET--"
				print "check indeed concrete"
				checkIndeedConcrete(fileorg)
			else:
				print "--BADOMIT-- need to refine"
				spuriousAns.append(interpretation)
				badomitAtoms.append(res)
				break
		print "------------------------------------------"
	#to avoid different answer sets with different bad omits for same abs I 
	#the badomits are minimized in the debug file
	if printatoms==True:
		text=""
		for a in atoms: 
			a=a.replace(" ","")
			text=text+"atom("+a+")."
		fout=open("outputs/org_program_atoms.lp","w")
		fout.write(text)
		fout.close()
		printatoms=False
	if len(badomitAtoms)==0:
		print "DONE"
		text=""
		for a in omitAtom: 
			a=a.replace(" ","")
			text=text+"omittedAtom("+a+")."
		fout=open("outputs/omitted_atoms.lp","w")
		fout.write(text)
		fout.close()
		break
	print badomitAtoms[0]
	addedBack.extend(badomitAtoms[0])

	print "REFINEMENT"
	refineStep=refineStep+1

	omitAtom = [s for s in omitAtom if s not in badomitAtoms[0]]
	
	print omitAtom
	if omitAtom==[]:
		print "refined all the way back - will not continue!"
		break
	print "------------------------------------------------"

	

	print "-------------------"
	

cR = createOmittedProgram(fileorg,omitAtom)
os.rename("outputs/omitted_program_info.lp", omitAtomFile+"_finalomit_info")
os.rename("outputs/omitted_program.lp", omitAtomFile+"_finalomit_absprogram")
if omitAtom!=[]:
	os.rename("outputs/omitted_atoms.lp", omitAtomFile+"_finalomittedatoms")
os.rename("outputs/org_program_atoms.lp", fileorg+"_progatoms")


fout=open(omitAtomFile+"_result_data","w")
text="TOTAL ATOM NUM: "+str(len(atoms))+"\n"
text=text+"ATOMS:"+" -".join(atoms)+"\n"
text=text+"TOTAL INITIAL OMIT NUM: "+str(len(omitAtomInit))+"\n"
text=text+"INITIAL OMIT ATOMS: "+" - ".join(omitAtomInit)+"\n"
text=text+"TOTAL ADDED BACK ATOM NUM: "+str(len(addedBack))+"\n"
text=text+"ADDED BACK ATOMS: "+ " - ".join(addedBack)+"\n"
text=text+"TOTAL REFINEMENT STEPS: "+str(refineStep)+"\n"
text=text+"TOTAL FINAL OMIT ATOM NUM: "+str(len(omitAtom))+"\n"
text=text+"FINAL OMIT ATOMS: "+" - ".join(omitAtom)+"\n"

fout.write(text)
fout.close()
